@@ -1,6 +1,6 @@
 /*
     GNOME Commander - A GNOME based file manager 
-    Copyright (C) 2001-2003 Marcus Bjurman
+    Copyright (C) 2001-2004 Marcus Bjurman
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -20,8 +20,8 @@
 #ifndef __CAP_H__
 #define __CAP_H__
 
-#include "gnome-cmd-dir-types.h"
 #include "gnome-cmd-file-list.h"
+#include "gnome-cmd-dir.h"
 
 void cap_cut_files (GnomeCmdFileList *fl, GList *files);
 void cap_copy_files (GnomeCmdFileList *fl, GList *files);
