@@ -1,6 +1,6 @@
 /*
     GNOME Commander - A GNOME based file manager 
-    Copyright (C) 2001-2003 Marcus Bjurman
+    Copyright (C) 2001-2004 Marcus Bjurman
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -19,8 +19,10 @@
 #include <config.h>
 #include "gnome-cmd-includes.h"
 #include <regex.h>
+#include "gnome-cmd-file-selector.h"
 #include "gnome-cmd-file-list.h"
-#include "gnome-cmd-file-funcs.h"
+#include "gnome-cmd-file.h"
+#include "gnome-cmd-main-win.h"
 #include "utils.h"
 #include "gnome-cmd-data.h"
 #include "gnome-cmd-patternsel-dialog.h"
@@ -34,6 +36,7 @@
 #include "gnome-cmd-delete-dialog.h"
 #include "gnome-cmd-advrename-dialog.h"
 #include "gnome-cmd-quicksearch-popup.h"
+#include "gnome-cmd-file-collection.h"
 #include "ls_colors.h"
 
 /* Controlls if file-uris should be escaped for local files when drag-N-dropping
@@ -109,7 +112,7 @@ struct _GnomeCmdFileListPrivate {
 	GnomeVFSListCompareFunc sort_func;
 	gint current_col;
 	gboolean sort_raising[FILE_LIST_NUM_COLUMNS];
-	GList *shown_files;    /* contains GnomeCmdFile pointers */
+	GnomeCmdFileCollection *shown_files;
 	GList *selected_files; /* contains GnomeCmdFile pointers */
 	gint cur_file;
 	gboolean shift_down;
@@ -122,6 +125,16 @@ struct _GnomeCmdFileListPrivate {
 };
 
 
+typedef struct
+{
+	gchar *text[FILE_LIST_NUM_COLUMNS];
+	
+	gchar *dpath;
+	gchar *fname;
+	gchar *fext;
+} FileFormatData;
+
+
 static int
 get_num_files (GnomeCmdFileList *fl)
 {
@@ -167,8 +180,6 @@ show_selpat_dialog (GnomeCmdFileList *fl, gboolean mode)
 	dialog = gnome_cmd_patternsel_dialog_new (fl, mode);
 
 	gtk_widget_ref (dialog);
-	gtk_object_set_data_full (GTK_OBJECT (fl), "gnome_cmd_patternsel_dialog", dialog,
-							  (GtkDestroyNotify)gtk_widget_unref);
 	gtk_signal_connect (GTK_OBJECT (dialog), "hide",
 						GTK_SIGNAL_FUNC (on_selpat_hide), fl);
 	gtk_widget_show (dialog);
@@ -196,8 +207,7 @@ on_quicksearch_popup_hide (GtkWidget *quicksearch_popup, GnomeCmdFileList *fl)
 
 
 static void
-do_select_file (GnomeCmdFileList *fl,
-				GnomeCmdFile *finfo)
+select_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo)
 {
 	gint row;
 	
@@ -215,9 +225,11 @@ do_select_file (GnomeCmdFileList *fl,
 	if (!gnome_cmd_data_get_use_ls_colors ())
 		gtk_clist_set_row_style (GTK_CLIST (fl), row, sel_list_style);
 	else {
-		GnomeCmdLayoutColors *colors = gnome_cmd_data_get_layout_colors ();
-		gtk_clist_set_foreground (GTK_CLIST (fl), row, colors->sel_fg);
-		gtk_clist_set_background (GTK_CLIST (fl), row, colors->sel_bg);
+		GnomeCmdColorTheme *colors = gnome_cmd_data_get_current_color_theme ();
+		if (!colors->respect_theme) {
+			gtk_clist_set_foreground (GTK_CLIST (fl), row, colors->sel_fg);
+			gtk_clist_set_background (GTK_CLIST (fl), row, colors->sel_bg);
+		}
 	}
 
 	if (g_list_index (fl->priv->selected_files, finfo) != -1)
@@ -232,8 +244,7 @@ do_select_file (GnomeCmdFileList *fl,
 
 
 static void
-do_unselect_file (GnomeCmdFileList *fl,
-				  GnomeCmdFile *finfo)
+unselect_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo)
 {
 	gint row;
 	
@@ -255,13 +266,13 @@ do_unselect_file (GnomeCmdFileList *fl,
 		gtk_clist_set_row_style (GTK_CLIST (fl), row, list_style);
 	else {
 		LsColor *col = ls_colors_get (finfo);
-		GnomeCmdLayoutColors *colors = gnome_cmd_data_get_layout_colors ();
+		GnomeCmdColorTheme *colors = gnome_cmd_data_get_current_color_theme ();
 		GdkColor *fg, *bg;
 		if (col) {
 			fg = col->fg;
 			bg = col->bg;
 			if (!fg) fg = colors->norm_fg;
-			if (!bg) bg = colors->norm_bg;			
+			if (!bg) bg = colors->norm_bg;
 			if (bg) gtk_clist_set_background (GTK_CLIST (fl), row, bg);
 			if (fg) gtk_clist_set_foreground (GTK_CLIST (fl), row, fg);
 		}
@@ -271,7 +282,7 @@ do_unselect_file (GnomeCmdFileList *fl,
 
 
 static void
-select_file (GnomeCmdFileList *fl,
+toggle_file (GnomeCmdFileList *fl,
 			 GnomeCmdFile *finfo)
 {
 	gint row;
@@ -283,19 +294,34 @@ select_file (GnomeCmdFileList *fl,
 	if (row == -1)
 		return;
 	
-	if (row < g_list_length (fl->priv->shown_files))
+	if (row < gnome_cmd_file_collection_get_size (fl->priv->shown_files))
 	{
 		if (g_list_index (fl->priv->selected_files, finfo) == -1)
-			do_select_file (fl, finfo);
+			select_file (fl, finfo);
 		else
-			do_unselect_file (fl, finfo);
+			unselect_file (fl, finfo);
 	}
 }
 
 
 static void
-select_file_at_row (GnomeCmdFileList *fl,
-					gint row)
+toggle_file_at_row (GnomeCmdFileList *fl, gint row)
+{
+	GnomeCmdFile *finfo;
+
+	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl));
+
+	fl->priv->cur_file = row;
+
+	finfo = get_file_at_row (fl, row);
+	if (!finfo) return;
+
+	toggle_file (fl, finfo);
+}
+
+
+static void
+select_file_at_row (GnomeCmdFileList *fl, gint row)
 {
 	GnomeCmdFile *finfo;
 
@@ -324,12 +350,32 @@ select_file_range (GnomeCmdFileList *fl,
 		start_row = end_row;
 		end_row = i;
 	}
-        
+
 	for ( i=start_row ; i<end_row+1 ; i++ )
 		select_file_at_row (fl, i);
 }
 
 
+static void
+toggle_file_range (GnomeCmdFileList *fl,
+				   gint start_row, gint end_row)
+{
+	gint i;
+
+	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl));
+
+	if (start_row > end_row)
+	{
+		i = start_row;
+		start_row = end_row;
+		end_row = i;
+	}
+
+	for ( i=start_row ; i<end_row+1 ; i++ )
+		toggle_file_at_row (fl, i);
+}
+
+
 static void toggle_files_with_same_extension (GnomeCmdFileList *fl,
 											  gboolean select)
 {
@@ -357,9 +403,9 @@ static void toggle_files_with_same_extension (GnomeCmdFileList *fl,
 
 			if (ext2 && strcmp (ext1, ext2) == 0) {
 				if (select)
-					do_select_file (fl, finfo);
+					select_file (fl, finfo);
 				else
-					do_unselect_file (fl, finfo);
+					unselect_file (fl, finfo);
 			}
 		}
 			
@@ -390,9 +436,9 @@ toggle_with_pattern (GnomeCmdFileList *fl, const gchar *pattern, gboolean case_s
 		if (finfo && finfo->info) {
 			if (filter_match (filter, finfo->info->name)) {
 				if (mode)
-					do_select_file (fl, finfo);
+					select_file (fl, finfo);
 				else
-					do_unselect_file (fl, finfo);
+					unselect_file (fl, finfo);
 			}
 		}
 			
@@ -431,7 +477,6 @@ create_column_titles (GnomeCmdFileList *fl)
 								  (GtkDestroyNotify) gtk_widget_unref);
 		gtk_widget_show (fl->priv->column_labels[i]);
 		gtk_box_pack_start (GTK_BOX (hbox), fl->priv->column_labels[i], TRUE, TRUE, 0);
-		gtk_widget_modify_style (fl->priv->column_labels[i], main_rc_style);
 
 		pixmap = gtk_pixmap_new (pm, bm);
 		gtk_widget_ref (pixmap);
@@ -442,10 +487,13 @@ create_column_titles (GnomeCmdFileList *fl)
 		
 		fl->priv->column_pixmaps[i] = pixmap;
 		gtk_clist_set_column_widget (GTK_CLIST (fl), i, hbox);
+	}
+
+	for ( i=1 ; i<FILE_LIST_NUM_COLUMNS ; i++ ) {
 		gtk_clist_set_column_width (GTK_CLIST (fl), i, gnome_cmd_data_get_fs_col_width (i));
 		gtk_clist_set_column_justification (GTK_CLIST (fl), i, list_column_justification[i]);
 	}
-
+	
 	gtk_clist_column_titles_show (GTK_CLIST (fl));	
 }
 
@@ -570,6 +618,8 @@ drag_data_get (GtkWidget        *widget,
 
 	data = build_selected_file_list (fl, &len);
 
+	if (!data) return;
+	
 	switch (info){
 		case TARGET_URI_LIST:
 		case TARGET_TEXT_PLAIN:
@@ -577,7 +627,7 @@ drag_data_get (GtkWidget        *widget,
 			break;
 
 		case TARGET_URL:
-			files = gnome_uri_list_extract_uris (data);
+			files = gnome_vfs_uri_list_parse (data);
 			if (files) {
 				gtk_selection_data_set (selection_data,
 										selection_data->target,
@@ -585,7 +635,7 @@ drag_data_get (GtkWidget        *widget,
 										files->data,
 										strlen (files->data));
 			}
-			gnome_uri_list_free_strings (files);
+			g_list_foreach (files, (GFunc)g_free, NULL);
 			break;
 
 		default:
@@ -645,7 +695,7 @@ show_file_popup (GnomeCmdFileList *fl, GdkEventButton *event)
 							  "file_popup_menu", menu,
 							  (GtkDestroyNotify)gtk_widget_unref);
 
-	gnome_popup_menu_do_popup (menu, NULL, fl, event, fl);
+	gnome_popup_menu_do_popup (menu, NULL, fl, event, fl, NULL);
 }
 
 
@@ -669,7 +719,7 @@ on_right_mb_timeout (GnomeCmdFileList *fl)
 	
 	if (fl->priv->right_mb_down_file == focus_file) {
 		show_file_popup (fl, NULL);
-		select_file (fl, focus_file);
+		toggle_file (fl, focus_file);
 		return FALSE;
 	}	
 
@@ -768,20 +818,31 @@ sort_by_extension (GnomeCmdFile *f1, GnomeCmdFile *f2, GnomeCmdFileList *fl)
 static gint
 sort_by_dir (GnomeCmdFile *f1, GnomeCmdFile *f2, GnomeCmdFileList *fl)
 {
+	gint ret = 0;
 	gboolean raising = fl->priv->sort_raising[fl->priv->current_col];
 
-	const gchar *d1 = gnome_cmd_file_get_dir_path (f1);
-	const gchar *d2 = gnome_cmd_file_get_dir_path (f2);
+	gchar *t1 = gnome_cmd_file_get_path (f1);
+	gchar *t2 = gnome_cmd_file_get_path (f2);
+	gchar *d1 = g_dirname (t1);
+	gchar *d2 = g_dirname (t2);
 
 	if (strcmp (f1->info->name, "..") == 0)
-		return -1;
+		ret = -1;
 	else if (strcmp (f2->info->name, "..") == 0)
-		return 1;
+		ret = 1;
 	else if (f1->info->type > f2->info->type)
-		return -1;
+		ret = -1;
 	else if (f1->info->type < f2->info->type)
-		return 1;
+		ret = 1;
 
+	g_free (t1);
+	g_free (t2);
+	g_free (d1);
+	g_free (d2);
+
+	if (ret)
+		return ret;
+	
 	return my_strcmp (d1, d2, raising);
 }
 
@@ -986,20 +1047,17 @@ on_scroll_vertical                  (GtkCList        *clist,
 		if (start_row < 0 || end_row < 0)
 			return;
         
-		if (end_row < num_files-1)
-			end_row--;
-		
         if (scroll_type == GTK_SCROLL_STEP_BACKWARD || scroll_type == GTK_SCROLL_STEP_FORWARD)
-			select_file_at_row (fl, start_row);
+			toggle_file_at_row (fl, start_row);
         else
-			select_file_range (fl, start_row, end_row);
+			toggle_file_range (fl, start_row, end_row);
     }
 
 	fl->priv->cur_file = clist->focus_row;
 }
 							   
 
-static gint
+static gboolean
 on_button_press (GtkCList *clist, GdkEventButton *event, GnomeCmdFileList *fl)
 {
 	gint row;
@@ -1015,7 +1073,7 @@ on_button_press (GtkCList *clist, GdkEventButton *event, GnomeCmdFileList *fl)
 	gtk_signal_emit (
 		GTK_OBJECT (fl), file_list_signals[LIST_CLICKED], event);
 	
-	row = gnome_cmd_file_list_get_row (fl, event->x, event->y);
+	row = gnome_cmd_clist_get_row (GNOME_CMD_CLIST (fl), event->x, event->y);
 	if (row < 0) {
 		gtk_signal_emit (
 			GTK_OBJECT (fl), file_list_signals[EMPTY_SPACE_CLICKED], event);
@@ -1073,17 +1131,17 @@ on_file_clicked (GnomeCmdFileList *fl,
 			if (event->state & GDK_SHIFT_MASK)
 				select_file_range (fl, fl->priv->shift_down_row, row);
 			else if (event->state & GDK_CONTROL_MASK)
-				select_file_at_row (fl, row);
+				toggle_file_at_row (fl, row);
 		}
 		else if (event->button == 3) {
 			if (strcmp (finfo->info->name, "..") != 0) {
 				if (gnome_cmd_data_get_right_mouse_button_mode() == RIGHT_BUTTON_SELECTS) {
 					if (g_list_index (fl->priv->selected_files, finfo) == -1) {
-						do_select_file (fl, finfo);
+						select_file (fl, finfo);
 						fl->priv->right_mouse_sel_state = 1;
 					}
 					else {
-						do_unselect_file (fl, finfo);
+						unselect_file (fl, finfo);
 						fl->priv->right_mouse_sel_state = 0;
 					}
 
@@ -1115,16 +1173,16 @@ on_motion_notify                     (GtkCList *clist,
 		y = event->y;
 		y -= (clist->column_title_area.height - GTK_CONTAINER (clist)->border_width);
 				
-		row = gnome_cmd_file_list_get_row (fl, event->x, y);
+		row = gnome_cmd_clist_get_row (GNOME_CMD_CLIST (fl), event->x, y);
 		
 		if (row != -1) {
 			GnomeCmdFile *finfo = gnome_cmd_file_list_get_file_at_row (fl, row+1);
 			if (finfo) {
 				gnome_cmd_file_list_select_row (fl, row+1);
 				if (fl->priv->right_mouse_sel_state)
-					do_select_file (fl, finfo);
+					select_file (fl, finfo);
 				else
-					do_unselect_file (fl, finfo);
+					unselect_file (fl, finfo);
 			}
 		}		
 	}
@@ -1143,7 +1201,7 @@ on_button_release (GtkWidget *widget, GdkEventButton *event, GnomeCmdFileList *f
 	if (GTK_CLIST (fl)->clist_window != event->window)
 		return FALSE;
 
-	row = gnome_cmd_file_list_get_row (fl, event->x, event->y);
+	row = gnome_cmd_clist_get_row (GNOME_CMD_CLIST (fl), event->x, event->y);
 	if (row < 0)
 		return FALSE;
 
@@ -1170,7 +1228,6 @@ on_realize                               (GnomeCmdFileList *fl,
 {
 	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl));
 
-	create_column_titles (fl);
 	update_column_sort_arrows (fl);
 }
 
@@ -1185,7 +1242,7 @@ destroy (GtkObject *object)
 {
 	GnomeCmdFileList *fl = GNOME_CMD_FILE_LIST (object);
 	
-	gnome_cmd_file_list_free (fl->priv->shown_files);
+	gtk_object_destroy (GTK_OBJECT (fl->priv->shown_files));
 	gnome_cmd_file_list_free (fl->priv->selected_files);
 
 	if (fl->priv)
@@ -1216,7 +1273,7 @@ class_init (GnomeCmdFileListClass *class)
 	file_list_signals[FILE_CLICKED] =
 		gtk_signal_new ("file_clicked",
 			GTK_RUN_LAST,
-		    object_class->type,
+		    G_OBJECT_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GnomeCmdFileListClass, file_clicked),
 		    gtk_marshal_NONE__POINTER_POINTER,
 		    GTK_TYPE_NONE,
@@ -1225,7 +1282,7 @@ class_init (GnomeCmdFileListClass *class)
 	file_list_signals[LIST_CLICKED] =
 		gtk_signal_new ("list_clicked",
 			GTK_RUN_LAST,
-		    object_class->type,
+		    G_OBJECT_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GnomeCmdFileListClass, list_clicked),
 		    gtk_marshal_NONE__POINTER,
 		    GTK_TYPE_NONE,
@@ -1234,7 +1291,7 @@ class_init (GnomeCmdFileListClass *class)
 	file_list_signals[EMPTY_SPACE_CLICKED] =
 		gtk_signal_new ("empty_space_clicked",
 			GTK_RUN_LAST,
-		    object_class->type,
+		    G_OBJECT_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GnomeCmdFileListClass, empty_space_clicked),
 		    gtk_marshal_NONE__POINTER,
 		    GTK_TYPE_NONE,
@@ -1243,14 +1300,12 @@ class_init (GnomeCmdFileListClass *class)
 	file_list_signals[SELECTION_CHANGED] =
 		gtk_signal_new ("selection_changed",
 			GTK_RUN_LAST,
-		    object_class->type,
+		    G_OBJECT_CLASS_TYPE (object_class),
 		    GTK_SIGNAL_OFFSET (GnomeCmdFileListClass, selection_changed),
 		    gtk_marshal_NONE__NONE,
 		    GTK_TYPE_NONE,
 			0);
 	
-	gtk_object_class_add_signals (object_class, file_list_signals, LAST_SIGNAL);
-	
 	object_class->destroy = destroy;
 	widget_class->map = map;	
 	class->file_clicked = NULL;
@@ -1264,7 +1319,7 @@ init (GnomeCmdFileList *fl)
 	gint i;
 	
 	fl->priv = g_new (GnomeCmdFileListPrivate, 1);
-	fl->priv->shown_files = NULL;
+	fl->priv->shown_files = gnome_cmd_file_collection_new ();
 	fl->priv->selected_files = NULL;
 	fl->priv->shift_down = FALSE;
 	fl->priv->sort_func = (GnomeVFSListCompareFunc)sort_by_name;
@@ -1274,23 +1329,24 @@ init (GnomeCmdFileList *fl)
 	fl->priv->right_mb_timeout_id = 0;
 	fl->priv->quicksearch_popup = NULL;
 	
-	list_column_titles[0] = "";
-	list_column_titles[1] = _("name");
-	list_column_titles[2] = _("ext");
-	list_column_titles[3] = _("dir");
-	list_column_titles[4] = _("size");
-	list_column_titles[5] = _("date");
-	list_column_titles[6] = _("perm");
-	list_column_titles[7] = _("uid");
-	list_column_titles[8] = _("gid");
+	list_column_titles[FILE_LIST_COLUMN_ICON]  = "";
+	list_column_titles[FILE_LIST_COLUMN_NAME]  = _("name");
+	list_column_titles[FILE_LIST_COLUMN_EXT]   = _("ext");
+	list_column_titles[FILE_LIST_COLUMN_DIR]   = _("dir");
+	list_column_titles[FILE_LIST_COLUMN_SIZE]  = _("size");
+	list_column_titles[FILE_LIST_COLUMN_DATE]  = _("date");
+	list_column_titles[FILE_LIST_COLUMN_PERM]  = _("perm");
+	list_column_titles[FILE_LIST_COLUMN_OWNER] = _("uid");
+	list_column_titles[FILE_LIST_COLUMN_GROUP] = _("gid");
 	
 	for ( i=0 ; i<FILE_LIST_NUM_COLUMNS ; i++ )
 		fl->priv->sort_raising[i] = FALSE;
 
-	gtk_clist_construct (GTK_CLIST (fl), FILE_LIST_NUM_COLUMNS, list_column_titles);
-	
 	init_dnd (fl);
 	
+	for ( i=0 ; i<FILE_LIST_NUM_COLUMNS ; i++ )
+		gtk_clist_set_column_resizeable (GTK_CLIST (fl), i, TRUE);
+	
 	gtk_signal_connect_after (GTK_OBJECT (fl), "scroll_vertical",
 							  GTK_SIGNAL_FUNC (on_scroll_vertical), fl);
 	gtk_signal_connect (GTK_OBJECT (fl), "click-column",
@@ -1303,7 +1359,7 @@ init (GnomeCmdFileList *fl)
 	gtk_signal_connect (GTK_OBJECT (fl), "motion-notify-event",
 						GTK_SIGNAL_FUNC (on_motion_notify), fl);
 	
-	gtk_signal_connect (GTK_OBJECT (fl), "realize",
+	gtk_signal_connect_after (GTK_OBJECT (fl), "realize",
 						GTK_SIGNAL_FUNC (on_realize), fl);
 	gtk_signal_connect (GTK_OBJECT (fl), "file-clicked",
 						GTK_SIGNAL_FUNC (on_file_clicked), fl);
@@ -1345,8 +1401,12 @@ gnome_cmd_file_list_new ()
 {
 	GnomeCmdFileList *fl;
 
-	fl = gtk_type_new (gnome_cmd_file_list_get_type ());
+	fl = g_object_new (gnome_cmd_file_list_get_type(),
+					   "n_columns", FILE_LIST_NUM_COLUMNS,
+					   NULL);
 
+	create_column_titles (fl);
+	
 	return GTK_WIDGET (fl);
 }
 
@@ -1372,64 +1432,106 @@ gnome_cmd_file_list_update_style (GnomeCmdFileList *fl)
 }
 
 
-
-/******************************************************************************
-*
-*   Function: gnome_cmd_file_list_add_file
-*
-*   Purpose:  Add a file to the list
-*
-*   Params:   @fl: The FileList to add the file to
-*             @finfo: The file to add
-*             @in_row: The row to add the file at. Set to -1 to append the file at the end.
-*
-*   Returns: 
-*
-*   Statuses: 
-*
-******************************************************************************/
-void
-gnome_cmd_file_list_add_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo, gint in_row)
+static gchar *
+strip_extension (const gchar *fname)
 {
-	gint row;
-	gchar *text[10];
-	GtkCList *clist;
+	gchar *s = g_strdup (fname);
+	gchar *p = g_strrstr (s, ".");
+	if (p && p != s)		
+		*p = '\0';
+	return s;
+}
 
-	clist = GTK_CLIST (fl);
 
+static void
+format_file_for_display (GnomeCmdFile *finfo, FileFormatData *data, gboolean tree_size)
+{
+	gchar *t, *t2;
+	gchar *size;
+
+	
 	/* If the user wants a character insted of icon for filetype set it now
 	 *
 	 */
 	if (gnome_cmd_data_get_layout () == GNOME_CMD_LAYOUT_TEXT)
-		text[0] = (gchar*)gnome_cmd_file_get_type_string (finfo);
+		data->text[FILE_LIST_COLUMN_ICON] = (gchar*)gnome_cmd_file_get_type_string (finfo);
 	else
-		text[0] = NULL;
+		data->text[FILE_LIST_COLUMN_ICON] = NULL;
 
+	/* Prepare the strings to show
+	 *
+	 */
+	t = gnome_cmd_file_get_path (finfo);
+	t2 = g_dirname (t);
+	g_free (t);
+	data->dpath = get_utf8 (t2);
+	g_free (t2);
+	
+	if (gnome_cmd_data_get_ext_disp_mode() == GNOME_CMD_EXT_DISP_STRIPPED
+		&& finfo->info->type == GNOME_VFS_FILE_TYPE_REGULAR) {
+		t = strip_extension (gnome_cmd_file_get_name (finfo));
+		data->fname = get_utf8 (t);
+		g_free (t);
+	}	
+	else
+		data->fname = get_utf8 (gnome_cmd_file_get_name (finfo) );
+
+	if (gnome_cmd_data_get_ext_disp_mode() != GNOME_CMD_EXT_DISP_WITH_FNAME)
+		data->fext = get_utf8 (gnome_cmd_file_get_extension (finfo));
+	else
+		data->fext = NULL;
+
+	if (tree_size)
+		size = (gchar*)gnome_cmd_file_get_tree_size (finfo);
+	else
+		size = (gchar*)gnome_cmd_file_get_size (finfo);
+	
 	/* Set other file information
 	 *
 	 */
-	text[1] = (gchar*)gnome_cmd_file_get_name (finfo);
-	text[2] = (gchar*)gnome_cmd_file_get_extension (finfo);
-	text[3] = (gchar*)gnome_cmd_file_get_dir_path (finfo);
-	text[4] = (gchar*)gnome_cmd_file_get_size (finfo);
-	text[5] = (gchar*)gnome_cmd_file_get_mdate (finfo, FALSE);
-	text[6] = (gchar*)gnome_cmd_file_get_perm (finfo);
-	text[7] = (gchar*)gnome_cmd_file_get_owner (finfo);
-	text[8] = (gchar*)gnome_cmd_file_get_group (finfo);
-	text[9] = NULL;
+	data->text[FILE_LIST_COLUMN_NAME]  = data->fname;
+ 	data->text[FILE_LIST_COLUMN_EXT]   = data->fext;
+	data->text[FILE_LIST_COLUMN_DIR]   = data->dpath;
+	data->text[FILE_LIST_COLUMN_SIZE]  = size;
+	data->text[FILE_LIST_COLUMN_DATE]  = (gchar*)gnome_cmd_file_get_mdate (finfo, FALSE);
+	data->text[FILE_LIST_COLUMN_PERM]  = (gchar*)gnome_cmd_file_get_perm (finfo);
+	data->text[FILE_LIST_COLUMN_OWNER] = (gchar*)gnome_cmd_file_get_owner (finfo);
+	data->text[FILE_LIST_COLUMN_GROUP] = (gchar*)gnome_cmd_file_get_group (finfo);
+	data->text[FILE_LIST_NUM_COLUMNS]  = NULL;
+}
 
-	/* Add the file to the list
-	 *
-	 */
-	gnome_cmd_file_ref (finfo);
-	fl->priv->shown_files = g_list_append (fl->priv->shown_files, finfo);	
+
+static void
+cleanup_file_format (FileFormatData *data)
+{
+	g_free (data->dpath);
+	g_free (data->fname);
+	if (data->fext)
+		g_free (data->fext);
+}
+
+
+static void
+add_file_to_clist (GnomeCmdFileList *fl, GnomeCmdFile *finfo, gint in_row)
+{
+	gint row;
+	GtkCList *clist;
+	FileFormatData data;
+
+	clist = GTK_CLIST (fl);
+
+	format_file_for_display (finfo, &data, FALSE);
+	
 	if (in_row == -1)
-		row = gtk_clist_append (clist, text);	
+		row = gtk_clist_append (clist, data.text);	
 	else
-		row = gtk_clist_insert (clist, in_row, text);	
+		row = gtk_clist_insert (clist, in_row, data.text);
+
+	cleanup_file_format (&data);
+
 	
 	/* Setup row data and color
-     *
+	 *
 	 */
 	if (!gnome_cmd_data_get_use_ls_colors ())
 		gtk_clist_set_row_style (clist, row, list_style);
@@ -1456,6 +1558,33 @@ gnome_cmd_file_list_add_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo, gint in
 }
 
 
+/******************************************************************************
+*
+*   Function: gnome_cmd_file_list_add_file
+*
+*   Purpose:  Add a file to the list
+*
+*   Params:   @fl: The FileList to add the file to
+*             @finfo: The file to add
+*             @in_row: The row to add the file at. Set to -1 to append the file at the end.
+*
+*   Returns: 
+*
+*   Statuses: 
+*
+******************************************************************************/
+void
+gnome_cmd_file_list_add_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo, gint row)
+{
+	/* Add the file to the list
+	 *
+	 */
+	gnome_cmd_file_collection_add (fl->priv->shown_files, finfo);
+
+	add_file_to_clist (fl, finfo, row);
+}
+
+
 /******************************************************************************
 *
 *   Function: gnome_cmd_file_list_show_files
@@ -1524,25 +1653,44 @@ gnome_cmd_file_list_insert_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo)
 void
 gnome_cmd_file_list_update_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo)
 {
-	gchar *text[10];
+	FileFormatData data;
 	gint i,row;
 
+	if (!gnome_cmd_file_needs_update (finfo))
+		return;
+	
 	row = get_row_from_file (fl, finfo);
 	if (row == -1)
 		return;
-	
-	text[1] = (gchar*)gnome_cmd_file_get_name (finfo);
-	text[2] = (gchar*)gnome_cmd_file_get_extension (finfo);
-	text[3] = (gchar*)gnome_cmd_file_get_dir_path (finfo);
-	text[4] = (gchar*)gnome_cmd_file_get_size (finfo);
-	text[5] = (gchar*)gnome_cmd_file_get_mdate (finfo, FALSE);
-	text[6] = (gchar*)gnome_cmd_file_get_perm (finfo);
-	text[7] = (gchar*)gnome_cmd_file_get_owner (finfo);
-	text[8] = (gchar*)gnome_cmd_file_get_group (finfo);
-	text[9] = NULL;
 
-	for ( i=1 ; i<9; i++ )
-		gtk_clist_set_text (GTK_CLIST (fl), row, i, text[i]);
+	format_file_for_display (finfo, &data, FALSE);
+
+	for ( i=1 ; i<FILE_LIST_NUM_COLUMNS; i++ )
+		gtk_clist_set_text (GTK_CLIST (fl), row, i, data.text[i]);
+
+	cleanup_file_format (&data);
+}
+
+
+void
+gnome_cmd_file_list_show_dir_size (GnomeCmdFileList *fl, GnomeCmdFile *finfo)
+{
+	gint i,row;
+	FileFormatData data;
+	
+	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl));
+	g_return_if_fail (GNOME_CMD_IS_FILE (finfo));
+	
+	row = get_row_from_file (fl, finfo);
+	if (row == -1)
+		return;
+
+	format_file_for_display (finfo, &data, TRUE);
+	
+	for ( i=1 ; i<FILE_LIST_NUM_COLUMNS; i++ )
+		gtk_clist_set_text (GTK_CLIST (fl), row, i, data.text[i]);
+
+	cleanup_file_format (&data);
 }
 
 
@@ -1555,12 +1703,28 @@ gnome_cmd_file_list_remove_file (GnomeCmdFileList *fl, GnomeCmdFile *finfo)
 	g_return_if_fail (finfo != NULL);
 	
 	row = get_row_from_file (fl, finfo);
-	gtk_clist_remove (GTK_CLIST (fl), row);
-	fl->priv->selected_files = g_list_remove (fl->priv->selected_files, finfo);
-	fl->priv->shown_files = g_list_remove (fl->priv->shown_files, finfo);
-	gnome_cmd_file_unref (finfo);
+	if (row >= 0) {
+		gtk_clist_remove (GTK_CLIST (fl), row);
+		fl->priv->selected_files = g_list_remove (fl->priv->selected_files, finfo);
+		gnome_cmd_file_collection_remove (fl->priv->shown_files, finfo);
 
-	focus_file_at_row (fl, MIN(row, GTK_CLIST(fl)->focus_row));
+		focus_file_at_row (fl, MIN(row, GTK_CLIST(fl)->focus_row));
+	}
+}
+
+
+void
+gnome_cmd_file_list_remove_file_by_uri (GnomeCmdFileList *fl, const gchar *uri_str)
+{
+	GnomeCmdFile *file;
+	
+	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl));
+	g_return_if_fail (uri_str != NULL);
+
+	file = gnome_cmd_file_collection_lookup (fl->priv->shown_files, uri_str);
+	g_return_if_fail (GNOME_CMD_IS_FILE (file));
+
+	gnome_cmd_file_list_remove_file (fl, file);
 }
 
 
@@ -1581,9 +1745,8 @@ gnome_cmd_file_list_remove_all_files (GnomeCmdFileList *fl)
 	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl));
 	
 	gtk_clist_clear (GTK_CLIST (fl));
-	gnome_cmd_file_list_free (fl->priv->shown_files);
+	gnome_cmd_file_collection_clear (fl->priv->shown_files);
 	gnome_cmd_file_list_free (fl->priv->selected_files);
-	fl->priv->shown_files = NULL;
 	fl->priv->selected_files = NULL;
 }
 
@@ -1673,7 +1836,7 @@ gnome_cmd_file_list_get_all_files (GnomeCmdFileList *fl)
 {
 	g_return_val_if_fail (GNOME_CMD_IS_FILE_LIST (fl), NULL);
 	
-	return fl->priv->shown_files;
+	return gnome_cmd_file_collection_get_list (fl->priv->shown_files);
 }
 
 
@@ -1752,7 +1915,7 @@ gnome_cmd_file_list_select_all (GnomeCmdFileList *fl)
 	tmp = gnome_cmd_file_list_get_all_files (fl);
 	while (tmp) {
 		GnomeCmdFile *finfo = (GnomeCmdFile*)tmp->data;
-		do_select_file (fl, finfo);
+		select_file (fl, finfo);
 		tmp = tmp->next;
 	}
 }
@@ -1768,7 +1931,7 @@ gnome_cmd_file_list_unselect_all (GnomeCmdFileList *fl)
 	tmp = g_list_copy (fl->priv->selected_files);
 	while (tmp)	{
 		GnomeCmdFile *finfo = (GnomeCmdFile*)tmp->data;
-		do_unselect_file (fl, finfo);
+		unselect_file (fl, finfo);
 		tmp = tmp->next;
 	}
 
@@ -1787,7 +1950,7 @@ gnome_cmd_file_list_toggle (GnomeCmdFileList *fl)
 
 	finfo = get_file_at_row (fl, fl->priv->cur_file);
 	if (finfo)
-		select_file (fl, finfo);
+		toggle_file (fl, finfo);
 }
 
 
@@ -1800,7 +1963,7 @@ gnome_cmd_file_list_toggle_and_step (GnomeCmdFileList *fl)
 
 	finfo = get_file_at_row (fl, fl->priv->cur_file);
 	if (finfo)
-		select_file (fl, finfo);
+		toggle_file (fl, finfo);
 	if (fl->priv->cur_file < get_num_files (fl)-1)
 		focus_file_at_row (fl, fl->priv->cur_file+1);
 }
@@ -1841,20 +2004,6 @@ gnome_cmd_file_list_focus_file (GnomeCmdFileList *fl,
 }
 
 
-gint
-gnome_cmd_file_list_get_row (GnomeCmdFileList *fl, gint x, gint y)
-{
-	gint row;
-
-	g_return_val_if_fail (GNOME_CMD_IS_FILE_LIST (fl), -1);
-	
-	if (gtk_clist_get_selection_info (GTK_CLIST (fl), x, y, &row, NULL) == 0)
-		row = -1;
-
-	return row;
-}
-
-
 void
 gnome_cmd_file_list_select_row (GnomeCmdFileList *fl, gint row)
 {
@@ -1896,9 +2045,9 @@ gnome_cmd_file_list_invert_selection (GnomeCmdFileList *fl)
 
 		if (finfo && finfo->info) {
 			if (g_list_index (sel, finfo) == -1)
-				do_select_file (fl, finfo);
+				select_file (fl, finfo);
 			else
-				do_unselect_file (fl, finfo);			
+				unselect_file (fl, finfo);			
 		}
 			
 		tmp = tmp->next;
@@ -1928,6 +2077,65 @@ gnome_cmd_file_list_restore_selection (GnomeCmdFileList *fl)
 }
 
 
+static gint 
+compare_filename(GnomeCmdFile *f1, GnomeCmdFile *f2)
+{
+	return strcmp(f1->info->name, f2->info->name);
+}
+
+
+void
+gnome_cmd_file_list_compare_directories (void)
+{
+	GnomeCmdFileSelector *fs1 = gnome_cmd_main_win_get_active_fs (main_win);
+	GnomeCmdFileList     *fl1 = fs1 ? fs1->list : NULL;
+	GnomeCmdFileSelector *fs2 = gnome_cmd_main_win_get_inactive_fs (main_win);
+	GnomeCmdFileList     *fl2 = fs2 ? fs2->list : NULL;
+
+	GList *i;
+	
+	g_return_if_fail (GNOME_CMD_IS_FILE_LIST (fl1) || GNOME_CMD_IS_FILE_LIST (fl2));
+	
+	gnome_cmd_file_list_unselect_all (fl1);
+	gnome_cmd_file_list_select_all (fl2);
+	
+	for (i=gnome_cmd_file_list_get_all_files(fl1); i; i=i->next)
+	{
+		GnomeCmdFile *f1 = (GnomeCmdFile*)i->data;
+		GnomeCmdFile *f2;
+		GList *gl2 = g_list_find_custom (fl2->priv->selected_files, f1, (GCompareFunc)compare_filename);
+		
+		if (!gl2)
+		{
+			select_file (fl1, f1);
+			continue;
+		}
+		
+		f2 = (GnomeCmdFile*)gl2->data;
+		
+		if (f1->info->type==GNOME_VFS_FILE_TYPE_DIRECTORY || f2->info->type==GNOME_VFS_FILE_TYPE_DIRECTORY)
+		{
+			unselect_file (fl2, f2);
+			continue;	
+		}
+		
+		if (f1->info->mtime > f2->info->mtime)
+		{
+			select_file (fl1, f1);
+			unselect_file (fl2, f2);
+			continue;	
+		}
+		
+		if (f1->info->mtime == f2->info->mtime) {
+			if (f1->info->size == f2->info->size)
+				unselect_file (fl2, f2);
+			else
+				select_file (fl1, f1);
+		}
+	}
+}
+
+
 void
 gnome_cmd_file_list_sort (GnomeCmdFileList *fl)
 {
@@ -1943,16 +2151,14 @@ gnome_cmd_file_list_sort (GnomeCmdFileList *fl)
 	gtk_clist_clear (GTK_CLIST (fl));
 
 	/* Resort the files and readd them to the list */
-	list = g_list_copy (gnome_cmd_file_list_get_all_files (fl));
+	list = gnome_cmd_file_list_get_all_files (fl);
 	list = gnome_vfs_list_sort (list, fl->priv->sort_func, fl);
 	while (list != NULL)
 	{
 		GnomeCmdFile *finfo;
 
 		finfo = GNOME_CMD_FILE (list->data);
-		fl->priv->shown_files = g_list_remove (
-			fl->priv->shown_files, finfo);
-		gnome_cmd_file_list_add_file (fl, finfo, -1);		
+		add_file_to_clist (fl, finfo, -1);
 		list = list->next;
 	}
 	g_list_free (list);
@@ -1970,7 +2176,7 @@ gnome_cmd_file_list_sort (GnomeCmdFileList *fl)
 		GnomeCmdFile *finfo;
 		
 		finfo = GNOME_CMD_FILE (list->data);
-		do_select_file (fl, finfo);
+		select_file (fl, finfo);
 		list = list->next;
 	}
 	
@@ -2005,8 +2211,6 @@ gnome_cmd_file_list_show_rename_dialog (GnomeCmdFileList *fl)
 		GtkWidget *dialog = gnome_cmd_rename_dialog_new (finfo);
 		
 		gtk_widget_ref (dialog);
-		gtk_object_set_data_full (GTK_OBJECT (fl), "gnome_cmd_rename_dialog", dialog,
-								  (GtkDestroyNotify)gtk_widget_unref);
 		gtk_widget_show (dialog);
 	}
 }
@@ -2042,8 +2246,6 @@ gnome_cmd_file_list_show_chown_dialog (GnomeCmdFileList *fl)
 	
 		dialog = gnome_cmd_chown_dialog_new (files);
 		gtk_widget_ref (dialog);
-		gtk_object_set_data_full (GTK_OBJECT (fl), "gnome_cmd_chown_dialog", dialog,
-								  (GtkDestroyNotify)gtk_widget_unref);
 		gtk_widget_show (dialog);
 		g_list_free (files);
 	}
@@ -2064,8 +2266,6 @@ gnome_cmd_file_list_show_chmod_dialog (GnomeCmdFileList *fl)
 	
 		dialog = gnome_cmd_chmod_dialog_new (files);
 		gtk_widget_ref (dialog);
-		gtk_object_set_data_full (GTK_OBJECT (fl), "gnome_cmd_chmod_dialog", dialog,
-								  (GtkDestroyNotify)gtk_widget_unref);
 		gtk_widget_show (dialog);
 		g_list_free (files);
 	}
@@ -2085,8 +2285,6 @@ gnome_cmd_file_list_show_advrename_dialog (GnomeCmdFileList *fl)
 		GtkWidget *dialog = gnome_cmd_advrename_dialog_new (files);
 
 		gtk_widget_ref (dialog);
-		gtk_object_set_data_full (GTK_OBJECT (fl), "gnome_cmd_advrename_dialog", dialog,
-								  (GtkDestroyNotify)gtk_widget_unref);
 		gtk_widget_show (dialog);
 	}
 }
@@ -2191,6 +2389,7 @@ gnome_cmd_file_list_show_quicksearch (GnomeCmdFileList *fl, gchar c)
 		GnomeCmdQuicksearchPopup *popup =
 			GNOME_CMD_QUICKSEARCH_POPUP (fl->priv->quicksearch_popup);
 		gtk_entry_set_text (GTK_ENTRY (popup->entry), text);
+		gtk_entry_set_position (GTK_ENTRY (popup->entry), 1);
 	}
 
 	gtk_signal_connect (GTK_OBJECT (fl->priv->quicksearch_popup), "hide",
@@ -2203,7 +2402,7 @@ gnome_cmd_file_list_keypressed (GnomeCmdFileList *fl,
 								GdkEventKey *event)
 {
 	gint num_files;
-	
+
 	g_return_val_if_fail (event != NULL, FALSE);
 	g_return_val_if_fail (GNOME_CMD_IS_FILE_LIST (fl), FALSE);
 
@@ -2223,10 +2422,6 @@ gnome_cmd_file_list_keypressed (GnomeCmdFileList *fl,
 			case GDK_KP_Subtract:
 				toggle_files_with_same_extension (fl, FALSE);
 				break;
-
-			case GDK_b: //FIXME: Remove this
-				g_printerr ("voffset: %d\n", GTK_CLIST(fl)->voffset);
-				break;
 		}
 	}
 	else if (state_is_ctrl_alt (event->state)) {
@@ -2237,61 +2432,60 @@ gnome_cmd_file_list_keypressed (GnomeCmdFileList *fl,
 	else if (state_is_shift (event->state)) {
 		switch (event->keyval)
 		{
+			case GDK_F2:
+				gnome_cmd_file_list_compare_directories ();
+				return TRUE;
+				
 			case GDK_F6:
 				gnome_cmd_file_list_show_rename_dialog (fl);
 				return TRUE;
 				
 			case GDK_Left:
 			case GDK_Right:
-				//FIXME: This doesn't work as expected
 				event->state -= GDK_SHIFT_MASK;
 				return FALSE;
 
 			case GDK_Page_Up:
 			case GDK_KP_Page_Up:
 				fl->priv->shift_down = TRUE;
-				event->state -= GDK_SHIFT_MASK;
-				event->keyval = GDK_Page_Up;
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_PAGE_BACKWARD, 0.0, NULL);
 				return FALSE;
 
 			case GDK_Page_Down:
 			case GDK_KP_Page_Down:
 				fl->priv->shift_down = TRUE;
-				event->state -= GDK_SHIFT_MASK;
-				event->keyval = GDK_Page_Down;
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_PAGE_FORWARD, 0.0, NULL);
 				return FALSE;
 
 			case GDK_Up:	
 			case GDK_KP_Up:
 				fl->priv->shift_down = TRUE;
-				event->state -= GDK_SHIFT_MASK;
-				event->keyval = GDK_Up;
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_STEP_BACKWARD, 0.0, NULL);
 				return FALSE;
 
 			case GDK_Down:
 			case GDK_KP_Down:
 				fl->priv->shift_down = TRUE;
-				event->state -= GDK_SHIFT_MASK;
-				event->keyval = GDK_Down;
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_STEP_FORWARD, 0.0, NULL);
 				return FALSE;
 			
 			case GDK_Home:
 			case GDK_KP_Home:
 				fl->priv->shift_down = TRUE;
-				GTK_CLIST (fl)->focus_row = 0;
-				on_scroll_vertical (GTK_CLIST (fl), GTK_SCROLL_JUMP, 0, fl);
-				focus_file_at_row (fl, 0);
-				gtk_clist_moveto (GTK_CLIST (fl), 0, 0, 0, 0);
-				return FALSE;
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_JUMP, 0.0);
+				return TRUE;
 				
 			case GDK_End:
 			case GDK_KP_End:
 				fl->priv->shift_down = TRUE;
-				GTK_CLIST (fl)->focus_row = num_files-1;
-				on_scroll_vertical (GTK_CLIST (fl), GTK_SCROLL_JUMP, 0, fl);
-				focus_file_at_row (fl, num_files-1);
-				gtk_clist_moveto (GTK_CLIST (fl), num_files-1, 0, 1, 0);
-				return FALSE;
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_JUMP, 1.0);
+				return TRUE;
 		}
 	}
 	else if (state_is_ctrl (event->state)) {
@@ -2361,7 +2555,9 @@ gnome_cmd_file_list_keypressed (GnomeCmdFileList *fl,
 				
 			case GDK_Insert:
 			case GDK_KP_Insert:
-				gnome_cmd_file_list_toggle_and_step (fl);
+				gnome_cmd_file_list_toggle (fl);
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_STEP_FORWARD, 0.0, NULL);
 				return TRUE;
 
 			case GDK_KP_Page_Up:
@@ -2382,16 +2578,14 @@ gnome_cmd_file_list_keypressed (GnomeCmdFileList *fl,
 			
 			case GDK_Home:
 			case GDK_KP_Home:
-				focus_file_at_row (fl, 0);
-				on_scroll_vertical (GTK_CLIST (fl), GTK_SCROLL_JUMP, 0, fl);
-				gtk_clist_moveto (GTK_CLIST (fl), 0, 0, 0, 0);
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_JUMP, 0.0);
 				return TRUE;
 
 			case GDK_End:
 			case GDK_KP_End:
-				focus_file_at_row (fl, num_files-1);
-				on_scroll_vertical (GTK_CLIST (fl), GTK_SCROLL_JUMP, 0, fl);
-				gtk_clist_moveto (GTK_CLIST (fl), num_files-1, 0, 1, 0);
+				gtk_signal_emit_by_name (
+					GTK_OBJECT (fl), "scroll_vertical", GTK_SCROLL_JUMP, 1.0);
 				return TRUE;
 
 			case GDK_Delete:
@@ -2479,5 +2673,3 @@ gnome_vfs_list_sort (GList *list,
 		 compare_func,
 		 data);
 }
-
-
